/*****************************************************************************
*
*    Copyright (C) 2005 Paul Dwerryhouse
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; version 2 of the License.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <config.h>
#include "packet.h"
#include "util.h"

#define PORT 3478

void error(const char *s)
{
    fprintf(stderr, "Error: %s\n", s);
    exit(1);
}

void usage(const char *s)
{
    fprintf(stderr, "Usage: %s [-c] [-d] [-u username] <hostname>\n", s);
    exit(1);
}

int main(int argc, char **argv)
{

    struct sockaddr_in c_addr, srv_addr;
    int s, c, n;
    struct hostent *he;
    unsigned char buf[8192];

    struct packet *p, *q;
    struct packet_s *ps;
    char *hostname,*username=0;
    unsigned short port = PORT;

    stun_tid_t tid;

    int change_ip = 0, change_port = 0;

    srandom(time(0));

    while ((c = getopt(argc, argv, "cdhu:")) >= 0) {
	switch (c) {
	case 'c':
	    change_ip = 1;
	    break;
	case 'd':
	    change_port = 1;
	    break;
	case 'u':
	    username = optarg;
	    break;
	case 'h':		/* Fall through */
	default:
	    usage(argv[0]);
	    break;		/* Never reached */
	}
    }

    if (argc - optind < 1)
	usage(argv[0]);

    hostname = argv[optind++];

    s = socket(PF_INET, SOCK_DGRAM, 0);

    if ((he = gethostbyname(hostname)) == NULL) {
	error("No such host.");
    }

    if (argc - optind > 0)
	port = atoi(argv[optind++]);

    srv_addr.sin_addr = *((struct in_addr *) he->h_addr);

    srv_addr.sin_family = AF_INET;
    srv_addr.sin_port = htons(port);
    memset(&(srv_addr.sin_zero), 0, 8);

    c_addr.sin_family = AF_INET;
    c_addr.sin_port = htons(7777);
    c_addr.sin_addr.s_addr = INADDR_ANY;
    memset(&(c_addr.sin_zero), 0, 8);

    create_random_tid(&tid);
    p = packet_create(STUN_MT_BINREQ, tid);

    if (change_ip || change_port) {
	add_change_request(p, change_ip, change_port);
    }

    /*
       add_response_address(p, 0, 7777);
       add_reflected_from(p, 0, 7777);
       add_changed_address(p, 0, 7777);
     */

    if (username && !add_username(p, username)) {
	error("Can't add username\n");
    }

    ps = serialise_packet(p);

    bind(s, (struct sockaddr *) &c_addr, sizeof(struct sockaddr));

    sendto(s, ps->data, ps->len, 0, (struct sockaddr *) &srv_addr,
	   sizeof(struct sockaddr));
    
    n = recvfrom(s, buf, 8192, 0, 0, 0);

    q = packet_parse(buf, n);

    printf("Response from: %s port %d\n", inet_ntoa(srv_addr.sin_addr),
	   ntohs(srv_addr.sin_port));

    switch (q->type) {
	    case STUN_MT_BINREQ:
		    printf("\tBinding Request\n");
		    break;
	    case STUN_MT_BINRES:
		    printf("\tBinding Response\n");
		    break;
	    case STUN_MT_BINERR:
		    printf("\tBinding Error Response\n");
		    break;
	    default:
		    break;
    }
		  

    if (q->mapped_address) {
	printf("\tMapped Address: %s port %d\n",
	       addr_ntoa(q->mapped_address->address),
	       q->mapped_address->port);
    }

    if (q->response_address) {
	printf("\tResponse Address: %s port %d\n",
	       addr_ntoa(q->response_address->address),
	       q->response_address->port);
    }

    if (q->source_address) {
	printf("\tSource Address: %s port %d\n",
	       addr_ntoa(q->source_address->address),
	       q->source_address->port);
    }

    if (q->changed_address) {
	printf("\tChanged Address: %s port %d\n",
	       addr_ntoa(q->changed_address->address),
	       q->changed_address->port);
    }

    if (q->reflected_from) {
	printf("\tReflected From: %s port %d\n",
	       addr_ntoa(q->reflected_from->address),
	       q->reflected_from->port);
    }

    close(s);

    exit(0);
}
