/*****************************************************************************
*
*    Copyright (C) 2005 Paul Dwerryhouse
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; version 2 of the License.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/select.h>
#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <config.h>
#include "packet.h"
#include "conf.h"

#define CONFIG_FILE SYSCONFDIR "/stunnerd.conf"

#define MAXBUF 8192

void usage(const char *s)
{
    fprintf(stderr, "Usage: %s [-d]\n", s);
    exit(1);
}

void cleanup(void);
void terminate(int);
void daemonise(int, char **);

int main(int argc, char **argv)
{

    int c, s[2][2], maxs = 0, i, j;
    int debug = 0;
    unsigned char buf[MAXBUF];
    fd_set fds;

    struct sockaddr_in srv_addr[2][2], c_addr;
    struct packet *p, *q;
    struct packet_s *qs;

    struct conf *config;

    while ((c = getopt(argc, argv, "d")) >= 0) {
	switch (c) {
	case 'd':
	    debug = 1;
	    break;
	case 'h':		/* Fall through */
	default:
	    usage(argv[0]);
	    break;		/* Never reached */
	}
    }

    if (!debug)
	daemonise(argc, argv);

    signal(SIGINT, terminate);	/* clean up after interrupt */
    signal(SIGTERM, terminate);	/* clean up after interrupt */
    atexit(cleanup);

    config = readconf(CONFIG_FILE);

    openlog(argv[0], LOG_NOWAIT | LOG_PID, LOG_DAEMON);

    FD_ZERO(&fds);

    for (j = 0; j < 2; j++) {
	for (i = 0; i < 2; i++) {

	    s[i][j] = socket(PF_INET, SOCK_DGRAM, 0);
	    if (maxs < s[i][j])
		maxs = s[i][j];

	    srv_addr[i][j].sin_family = AF_INET;
	    srv_addr[i][j].sin_port = htons(config->port[j]);
	    srv_addr[i][j].sin_addr.s_addr = config->addr[i];
	    memset(&(srv_addr[i][j].sin_zero), 0, 8);

	    bind(s[i][j], (struct sockaddr *) &(srv_addr[i][j]),
		 sizeof(struct sockaddr));

	    FD_SET(s[i][j], &fds);
	}
    }

    syslog(LOG_INFO, "started.");

    while ((c = select(maxs + 1, &fds, NULL, NULL, NULL))) {

	unsigned int len = sizeof(struct sockaddr);

	for (j = 0; j < 2; j++) {
	    for (i = 0; i < 2; i++) {

		if (FD_ISSET(s[i][j], &fds)) {
		    int n, fd;

		    fd = s[i][j];

		    n = recvfrom(fd, buf, MAXBUF, 0,
				 (struct sockaddr *) &c_addr, &len);

		    printf("Received %d bytes from: %s port %d\n", n,
			   inet_ntoa(c_addr.sin_addr),
			   ntohs(c_addr.sin_port));

		    syslog(LOG_INFO, "Packet received from: %s "
			   "port %d\n",
			   inet_ntoa(c_addr.sin_addr),
			   ntohs(c_addr.sin_port));

		    p = packet_parse(buf, n);

		    fprintf(stderr,"x: %d %d\n",n, p == 0);
		    if (!p) {
			stun_tid_t tid;
			create_random_tid(&tid);

		    fprintf(stderr,"x: %d %d\n",n, p == 0);
			q = packet_create(STUN_MT_BINERR, tid);
		    fprintf(stderr,"x: %d %d\n",n, p == 0);
			add_error_code(q, STUN_ERR_BAD_REQ);
		    fprintf(stderr,"x: %d %d\n",n, p == 0);
		    }

		    else if (p->unknown_attributes) {
			int i;

			q = packet_create(STUN_MT_BINERR, p->tid);


			for (i = 0; i < p->unknown_attributes->n; i++) {

			    add_unknown_attribute(q,
						  ntohs(p->
							unknown_attributes->
							attr[i]));
			}

			add_error_code(q, STUN_ERR_UNKNOWN_ATTR);
		    }

		    else {
			q = packet_create(STUN_MT_BINRES, p->tid);

			add_mapped_address(q,
					   ntohl(c_addr.sin_addr.s_addr),
					   ntohs(c_addr.sin_port));

			add_source_address(q, ntohl(config->addr[i]),
					   config->port[j]);

			add_changed_address(q, ntohl(config->addr[1 - i]),
					    config->port[1 - j]);
		    }

		    qs = serialise_packet(q);

		    sendto(fd, qs->data, qs->len, 0,
			   (struct sockaddr *) &c_addr,
			   sizeof(struct sockaddr));

		    if (p) destroy_packet(p);
		    if (q) destroy_packet(q);
		    if (qs) destroy_packet_s(qs);

		}

		FD_SET(s[i][j], &fds);
	    }
	}
    }

    cleanup();
    exit(0);
}

void cleanup(void)
{
    syslog(LOG_INFO, "stopping.");
    closelog();
}

void terminate(int x)
{
    cleanup();
    exit(0);
}


void daemonise(int argc, char **argv)
{

    pid_t pid;

    if ((pid = fork()) < 0) {
	syslog(LOG_ERR, "Can't fork. Exiting.");
	exit(1);
    }

    if (pid)
	exit(0);

    /* Child */

    setsid();

    if ((pid = fork()) < 0) {
	syslog(LOG_ERR, "Can't fork. Exiting.");
	exit(1);
    }

    if (pid)
	exit(0);

    chdir("/");

    close(0);
    close(1);
    close(2);

}
